import random

from .utils import ExecuteCommand, ExecuteCommandWithOutput

def CheckForwarding() -> None:
    cmd = "cat /proc/sys/net/ipv4/ip_forward"
    if ExecuteCommandWithOutput(cmd) != "1":
        ExecuteCommand("sudo sh -c 'echo \"net.ipv4.ip_forward=1\" >> /etc/sysctl.conf'")

    cmd = "cat /proc/sys/net/ipv4/conf/all/forwarding"
    if ExecuteCommandWithOutput(cmd) != "1":
        ExecuteCommand("sudo sh -c 'echo \"net.ipv4.conf.all.forwarding=1\" >> /etc/sysctl.conf'")

    cmd = "cat /proc/sys/net/ipv6/conf/all/forwarding"
    if ExecuteCommandWithOutput(cmd) != "1":
        ExecuteCommand("sudo sh -c 'echo \"net.ipv6.conf.all.forwarding=1\" >> /etc/sysctl.conf'")

    ExecuteCommand("sudo sysctl -p")


def GetMacAddress() -> str:
    mac = "52:54:00:%02x:%02x:%02x" % (
            random.randint(0, 255),
            random.randint(0, 255),
            random.randint(0, 255))

    return mac

def AddLinuxBridge(name: str, ip4: str, ip6: str, plen4: int, plen6: int) -> None:
    mac_address = GetMacAddress()
    dummy_intf = "{}-nic".format(name)
    ExecuteCommand("sudo ip link add {} address {} type dummy".format(dummy_intf, mac_address))
    ExecuteCommand("sudo brctl addbr {}".format(name))
    ExecuteCommand("sudo brctl stp {} on".format(name))
    ExecuteCommand("sudo brctl addif {} {}".format(name, dummy_intf))
    ExecuteCommand("sudo ip address add {}/{} dev {}".format(ip4, plen4, name))
    if ip6 != "None":
        ExecuteCommand("sudo ip -6 address add {}/{} dev {}".format(ip6, plen6, name))
    ExecuteCommand("sudo ip link set dev {} up".format(dummy_intf))
    ExecuteCommand("sudo ip link set dev {} up".format(name))

def AddIptableRules(name: str, nw4: str) -> None:
    ExecuteCommand("sudo iptables -t mangle -A POSTROUTING "
                   "-o {} -p udp -m udp --dport 68 -j CHECKSUM --checksum-fill".format(name))
    ExecuteCommand("sudo iptables -t nat -A POSTROUTING "
                   "-s {} -d 224.0.0.0/24 -j RETURN".format(nw4))
    ExecuteCommand("sudo iptables -t nat -A POSTROUTING "
                   "-s {} -d 255.255.255.255/32 -j RETURN".format(nw4))
    ExecuteCommand("sudo iptables -t nat -A POSTROUTING "
                   "-s {} ! -d {} -p tcp -j MASQUERADE --to-ports 1024-65535".format(nw4, nw4))
    ExecuteCommand("sudo iptables -t nat -A POSTROUTING "
                   "-s {} ! -d {} -p udp -j MASQUERADE --to-ports 1024-65535".format(nw4, nw4))
    ExecuteCommand("sudo iptables -t nat -A POSTROUTING -s {} ! -d {} -j MASQUERADE".format(nw4, nw4))
    ExecuteCommand("sudo iptables -t filter -A INPUT "
                   "-i {} -p udp -m udp -m multiport --dports 53,67 -j ACCEPT".format(name))
    ExecuteCommand("sudo iptables -t filter -A INPUT -i "
                   "{} -p tcp -m tcp -m multiport --dports 53,67 -j ACCEPT".format(name))
    ExecuteCommand("sudo iptables -t filter -A FORWARD -d {} -o {} -j ACCEPT".format(nw4, name))
    ExecuteCommand("sudo iptables -t filter -A FORWARD -s {} -i {} -j ACCEPT".format(nw4, name))
    ExecuteCommand("sudo iptables -t filter -A FORWARD -i {} -j ACCEPT".format(name))
    ExecuteCommand("sudo iptables -t filter -A FORWARD -o {} -j ACCEPT".format(name))
    ExecuteCommand("sudo sh -c 'iptables-save > /etc/iptables/rules.v4'")

def DelLinuxBridge(name: str) -> None:
    dummy_intf = "{}-nic".format(name)
    ExecuteCommand("sudo ip link set dev {} down".format(dummy_intf))
    ExecuteCommand("sudo ip link set dev {} down".format(name))
    ExecuteCommand("sudo ip link del {}".format(dummy_intf))
    ExecuteCommand("sudo ip link del {}".format(name))

def DelIptableRules(name: str, nw4: str) -> None:
    ExecuteCommand("sudo iptables -t mangle -D POSTROUTING "
                   "-o {} -p udp -m udp --dport 68 -j CHECKSUM --checksum-fill".format(name))
    ExecuteCommand("sudo iptables -t nat -D POSTROUTING "
                   "-s {} -d 224.0.0.0/24 -j RETURN".format(nw4))
    ExecuteCommand("sudo iptables -t nat -D POSTROUTING "
                   "-s {} -d 255.255.255.255/32 -j RETURN".format(nw4))
    ExecuteCommand("sudo iptables -t nat -D POSTROUTING "
                   "-s {} ! -d {} -p tcp -j MASQUERADE --to-ports 1024-65535".format(nw4, nw4))
    ExecuteCommand("sudo iptables -t nat -D POSTROUTING "
                   "-s {} ! -d {} -p udp -j MASQUERADE --to-ports 1024-65535".format(nw4, nw4))
    ExecuteCommand("sudo iptables -t nat -D POSTROUTING -s {} ! -d {} -j MASQUERADE".format(nw4, nw4))
    ExecuteCommand("sudo iptables -t filter -D INPUT "
                   "-i {} -p udp -m udp -m multiport --dports 53,67 -j ACCEPT".format(name))
    ExecuteCommand("sudo iptables -t filter -D INPUT -i "
                   "{} -p tcp -m tcp -m multiport --dports 53,67 -j ACCEPT".format(name))
    ExecuteCommand("sudo iptables -t filter -D FORWARD -d {} -o {} -j ACCEPT".format(nw4, name))
    ExecuteCommand("sudo iptables -t filter -D FORWARD -s {} -i {} -j ACCEPT".format(nw4, name))
    ExecuteCommand("sudo iptables -t filter -D FORWARD -i {} -j ACCEPT".format(name))
    ExecuteCommand("sudo iptables -t filter -D FORWARD -o {} -j ACCEPT".format(name))
    ExecuteCommand("sudo sh -c 'iptables-save > /etc/iptables/rules.v4'")



import json
import logging
import sys

import deployer.globals as G
from deployer.network import Network
from deployer.topology import Topology
from deployer.utils import ProcessArguments
from deployer.virtual_machine import VirtualMachine

class Deployer:
    config_ = {}
    topology_ = Topology()

    def __init__(self, config:str):
        with open(config) as conf_file:
            self.config_ = json.load(conf_file)

        Deployer.__validate_version(self.config_)

        self.__parse_networks()
        self.__parse_vms()
    #end __init__

    @staticmethod
    def __validate_version(config:dict):
        if config is None:
            logging.critical(f"Error parsing config file {config}")
            sys.exit(1)

        if config.get("version", None) is None:
            logging.error("This looks like older version of config which is "
                          " supported only by 0.0.x version")
            sys.exit(1)

        if config["version"] != 2:
            logging.error("This version supports only version 2 of config")
            sys.exit(1)
    #end __validate_version

    def __parse_networks(self):
        nw_configs = self.config_.get("networks", None)

        if nw_configs is None:
            logging.critical("No network config found. Exiting")
            sys.exit(1)

        if not isinstance(nw_configs, list):
            logging.critical("'network' config should be a list")
            sys.exit(1)

        for nw_conf in nw_configs:
            name = nw_conf["name"]
            nw = Network(nw_conf)
            self.Topology().AddNetwork(name, nw)
    #end __parse_networks

    def __parse_vms(self):
        vm_configs = self.config_.get("vms", None)

        if vm_configs is None:
            logging.critical("No VMs config found. Exiting")
            sys.exit(1)

        if not isinstance(vm_configs, list):
            logging.critical("'vms' config should be a list")
            sys.exit(1)

        for vm_conf in vm_configs:
            name = vm_conf["name"]
            vm = VirtualMachine(vm_conf)
            self.Topology().AddVm(name, vm)
    #end __parse_vms

    def Topology(self):
        return self.topology_
    #end Topology

#end Deployer

def topology_deployer():
    conf_file = ProcessArguments()
    tDeployer = Deployer(conf_file)
    topology = tDeployer.Topology()

    if G.OP == G.OP_CREATE:
        topology.Create()
    elif G.OP == G.OP_DELETE:
        topology.Delete()
    else:
        logging.critical(f"Invalid operation {G.OP}")
        sys.exit(1)
